import collections.abc
import dataclasses

import typing_extensions as typing

from ..constraints import (
    AllowedTypes,
    Constraint,
    ContentType,
    ElementCount,
    FullyQualifiedIdentifier,
    Length,
    MaximalElementCount,
    MaximalExclusive,
    MaximalInclusive,
    MaximalLength,
    MinimalElementCount,
    MinimalExclusive,
    MinimalInclusive,
    MinimalLength,
    Pattern,
    Schema,
    Set,
    Unit,
)
from ..fdl import Characters, Deserializer, EndElement, Serializer, StartElement
from ..protobuf import Reader, Writer
from .any import Any
from .convertable import Native
from .data_type import BasicType, DataType

if typing.TYPE_CHECKING:
    from ..common import Context
    from .list import List
    from .void import Void

T = typing.TypeVar("T", bound=Native)


@dataclasses.dataclass
class Constrained(DataType[T]):
    """
    a SiLA basic or list type with one or more constraints.

    Multiple constraints act together as a logical conjunction (and).

    Attributes:
      data_type: The SiLA data type of the constrained value.
      constraints: The list of constraints that must be satisfied by
        the constrained value.
      value: The SiLA data type instance.
    """

    data_type: typing.ClassVar[type[typing.Union[BasicType, "List"]]] = Any
    constraints: typing.ClassVar[collections.abc.Sequence["Constraint"]] = []

    value: typing.Union[BasicType, "List"] = dataclasses.field(default_factory=Any)

    @typing.override
    @classmethod
    async def from_native(cls, context: "Context", value: typing.Optional[T] = None, /) -> typing.Self:
        for constraint in cls.constraints:
            await constraint.validate(value)

        return await cls(value=await cls.data_type.from_native(context, value)).validate()

    @typing.override
    async def to_native(self, context: "Context", /) -> T:
        await self.validate()

        value = await self.value.to_native(context)

        for constraint in self.constraints:
            await constraint.validate(value)

        return value

    @typing.override
    async def validate(self) -> typing.Self:
        if not isinstance(self.value, self.data_type):
            msg = f"Expected value of type '{self.data_type.__name__}', received '{type(self.value).__name__}'."
            raise TypeError(msg)

        return self

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        from .list import List

        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = cls()
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            if isinstance(message.value, List):
                if not message.value.value:
                    message.value = message.data_type.decode(reader, length)
                    break
            else:
                message.value = message.data_type.decode(reader, length)

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        self.value.encode(writer, number)

        return writer.finish()

    @typing.override
    @classmethod
    def serialize(cls, serializer: Serializer) -> None:
        serializer.start_element("DataType")
        serializer.start_element("Constrained")
        cls.data_type.serialize(serializer)
        serializer.start_element("Constraints")
        for constraint in cls.constraints:
            constraint.serialize(serializer)
        serializer.end_element("Constraints")
        serializer.end_element("Constrained")
        serializer.end_element("DataType")

    @typing.override
    @classmethod
    def deserialize(
        cls, deserializer: Deserializer
    ) -> collections.abc.Generator[None, typing.Any, typing.Union[type["Void"], type["Constrained"]]]:
        from .list import List
        from .string import String
        from .void import Void

        yield from deserializer.read_start_element(name="Constrained")

        data_type = yield from deserializer.read(DataType.deserialize(deserializer))

        if not issubclass(data_type, (BasicType, List)):
            msg = f"Expected constraint's data type to be 'BasicType' or 'List', received '{data_type.__name__}'."
            raise ValueError(msg)

        yield from deserializer.read_start_element(name="Constraints")

        constraints: list[Constraint] = []
        while True:
            token = yield from deserializer.peek()

            if isinstance(token, StartElement):
                if token.name == "Length":
                    constraint = yield from deserializer.read(Length.deserialize(deserializer, data_type))
                    constraints.append(constraint)
                elif token.name == "MinimalLength":
                    constraint = yield from deserializer.read(MinimalLength.deserialize(deserializer, data_type))
                    constraints.append(constraint)
                elif token.name == "MaximalLength":
                    constraint = yield from deserializer.read(MaximalLength.deserialize(deserializer, data_type))
                    constraints.append(constraint)
                elif token.name == "Set":
                    constraint = yield from deserializer.read(Set.deserialize(deserializer, data_type))
                    constraints.append(constraint)
                elif token.name == "Pattern":
                    constraint = yield from deserializer.read(Pattern.deserialize(deserializer, data_type))
                    constraints.append(constraint)
                elif token.name == "MaximalExclusive":
                    constraint = yield from deserializer.read(MaximalExclusive.deserialize(deserializer, data_type))
                    constraints.append(constraint)
                elif token.name == "MaximalInclusive":
                    constraint = yield from deserializer.read(MaximalInclusive.deserialize(deserializer, data_type))
                    constraints.append(constraint)
                elif token.name == "MinimalExclusive":
                    constraint = yield from deserializer.read(MinimalExclusive.deserialize(deserializer, data_type))
                    constraints.append(constraint)
                elif token.name == "MinimalInclusive":
                    constraint = yield from deserializer.read(MinimalInclusive.deserialize(deserializer, data_type))
                    constraints.append(constraint)
                elif token.name == "Unit":
                    constraint = yield from deserializer.read(Unit.deserialize(deserializer, data_type))
                    constraints.append(constraint)
                elif token.name == "ContentType":
                    constraint = yield from deserializer.read(ContentType.deserialize(deserializer, data_type))
                    constraints.append(constraint)
                elif token.name == "ElementCount":
                    constraint = yield from deserializer.read(ElementCount.deserialize(deserializer, data_type))
                    constraints.append(constraint)
                elif token.name == "MinimalElementCount":
                    constraint = yield from deserializer.read(MinimalElementCount.deserialize(deserializer, data_type))
                    constraints.append(constraint)
                elif token.name == "MaximalElementCount":
                    constraint = yield from deserializer.read(MaximalElementCount.deserialize(deserializer, data_type))
                    constraints.append(constraint)
                elif token.name == "FullyQualifiedIdentifier":
                    constraint = yield from deserializer.read(
                        FullyQualifiedIdentifier.deserialize(deserializer, data_type)
                    )
                    constraints.append(constraint)
                elif token.name == "Schema":
                    constraint = yield from deserializer.read(Schema.deserialize(deserializer, data_type))
                    constraints.append(constraint)
                elif token.name == "AllowedTypes":
                    constraint = yield from deserializer.read(AllowedTypes.deserialize(deserializer, data_type))
                    constraints.append(constraint)
                else:
                    msg = (
                        f"Expected start element with a constraint name, "
                        f"received start element with name '{token.name}'."
                    )
                    raise ValueError(msg)

            elif isinstance(token, EndElement):
                if token.name == "Constraints":
                    yield from deserializer.read_end_element("Constraints")
                    break
                else:
                    break  # pragma: no cover

            elif isinstance(token, Characters):
                msg = f"Expected start element with a constraint name, received characters '{token.value}'."
                raise ValueError(msg)

        yield from deserializer.read_end_element(name="Constrained")

        if data_type == String and constraints == [Length(0)]:
            return Void

        return cls.create(data_type, constraints)

    @classmethod
    def create(
        cls,
        data_type: type[typing.Union[BasicType, "List"]] = Any,
        constraints: typing.Optional[collections.abc.Sequence["Constraint"]] = None,
        name: str = "",
    ) -> type[typing.Self]:
        """
        Create a new SiLA `Constrained` class with the provided data type and constraints.

        Args:
          data_type: The SiLA data type of the constrained value.
          constraints: A sequence of constraints to apply to the data
            type.
          name: An optional name for the new `Constrained` class.

        Returns:
          A new `Constrained` class with the specified data type.
        """

        return dataclasses.make_dataclass(
            name or cls.__name__,
            [("value", data_type, dataclasses.field(default_factory=data_type))],
            bases=(cls,),
            namespace={"data_type": data_type, "constraints": constraints or []},
            eq=False,
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Constrained):
            return NotImplemented

        return (
            self.data_type.__name__ == other.data_type.__name__
            and self.constraints == other.constraints
            and self.value == other.value
        )



import typing_extensions as typing

if typing.TYPE_CHECKING:
    from .reader import Reader
    from .writer import Writer


class Message(typing.Protocol):
    """Protocol for encoding and decoding Protocol Buffer messages."""

    @classmethod
    def decode(cls, reader: typing.Union["Reader", bytes], length: typing.Optional[int] = None) -> typing.Self:
        """
        Decode a message instance from a protobuf byte stream.

        Args:
          reader: The data source from which to decode the message.
          length: An optional length argument that indicates how much
            data to decode. If not provided, the entire message will be
            decoded.

        Returns:
          An instance of the decoded message.

        Raises:
          DecodeError: If there is an error while decoding the message.
        """
        ...

    def encode(self, writer: typing.Optional["Writer"] = None, number: typing.Optional[int] = None) -> bytes:
        """
        Encode the message instance into a protobuf byte stream.

        Args:
          writer: The data sink to which the encoded message will be
            written.
          number: An optional field number to associate with the message
            during encoding.

        Returns:
          The encoded message as bytes.

        Raises:
          EncodeError: If there is an error while encoding the message.
        """
        ...



from .allowed_types import AllowedTypes
from .constraint import Constraint
from .content_type import ContentType, ContentTypeParameter
from .element_count import ElementCount
from .fully_qualified_identifier import FullyQualifiedIdentifier
from .length import Length
from .maximal_element_count import MaximalElementCount
from .maximal_exclusive import MaximalExclusive
from .maximal_inclusive import MaximalInclusive
from .maximal_length import MaximalLength
from .minimal_element_count import MinimalElementCount
from .minimal_exclusive import MinimalExclusive
from .minimal_inclusive import MinimalInclusive
from .minimal_length import MinimalLength
from .pattern import Pattern
from .schema import Schema, SchemaType
from .set import Set
from .unit import SIUnit, Unit, UnitComponent

__all__ = [
    "AllowedTypes",
    "Constraint",
    "ContentType",
    "ContentTypeParameter",
    "ElementCount",
    "FullyQualifiedIdentifier",
    "Length",
    "MaximalElementCount",
    "MaximalExclusive",
    "MaximalInclusive",
    "MaximalLength",
    "MinimalElementCount",
    "MinimalExclusive",
    "MinimalInclusive",
    "MinimalLength",
    "Pattern",
    "SIUnit",
    "Schema",
    "SchemaType",
    "Set",
    "Unit",
    "UnitComponent",
]



import collections.abc

import typing_extensions as typing

if typing.TYPE_CHECKING:
    from .deserializer import Deserializer
    from .serializer import Serializer


class Serializable(typing.Protocol):
    """Serialize this class into its xml representation and back."""

    @classmethod
    def serialize(cls, serializer: "Serializer") -> None:
        """
        Serialize the SiLA data type into the xml-based feature language definition.

        Args:
          serializer: The serializer instance used to write xml tokens.
        """
        ...

    @classmethod
    def deserialize(
        cls, deserializer: "Deserializer"
    ) -> collections.abc.Generator[None, typing.Any, type[typing.Self]]:
        """
        Deserialize the xml-based feature language definition into SiLA data types.

        Args:
          deserializer: The deserializer instance used to read xml
            tokens.

        Raises:
          ValueError: If an invalid or unexpected token is detected.
        """
        ...



import enum


class WireType(enum.IntEnum):
    """The wire type tells the parser how big the payload after it is."""

    VARINT = 0
    """
    Used for int32, int64, uint32, uint64, sint32, sint64, bool and
    enum.
    """

    I64 = 1
    """
    Used for fixed64, sfixed64 and double.
    """

    LEN = 2
    """
    Used for string, bytes, embedded messages and packed repeated
    fields.
    """

    SGROUP = 3
    """
    A deprecated wire type to indicate the start of a group.
    """

    EGROUP = 4
    """
    A deprecated wire type to indicate the end of a group.
    """

    I32 = 5
    """
    Used for fixed32, sfixed32 and float.
    """



import dataclasses

import typing_extensions as typing

from ..protobuf import Message, Reader, WireType, Writer


@dataclasses.dataclass
class CreateBinaryRequest(Message):
    """
    Initiate the process of transferring binary data in chunks.

    Attributes:
      binary_size: The total size of the binary data to be
        transferred, in bytes.
      chunk_count: The total number of chunks that the binary data
        will be divided into for transfer.
      parameter_identifier: A unique identifier for the parameter
        associated with this binary data transfer.
    """

    binary_size: int = 0
    chunk_count: int = 0
    parameter_identifier: str = ""

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = cls()
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.VARINT)
                message.binary_size = reader.read_uint64()
            elif field_number == 2:
                reader.expect_type(tag, WireType.VARINT)
                message.chunk_count = reader.read_uint32()
            elif field_number == 3:
                reader.expect_type(tag, WireType.LEN)
                message.parameter_identifier = reader.read_string()
            else:
                reader.skip_type(tag & 7)

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        if self.binary_size:
            writer.write_uint32(8).write_uint64(self.binary_size)
        if self.chunk_count:
            writer.write_uint32(16).write_uint32(self.chunk_count)
        if self.parameter_identifier:
            writer.write_uint32(26).write_string(self.parameter_identifier)

        if number:
            writer.ldelim()

        return writer.finish()



import collections.abc
import dataclasses

import typing_extensions as typing

from ..fdl import Deserializer, Serializer
from ..protobuf import ConversionError, DecodeError, Reader, WireType, Writer
from .any import Any
from .convertable import Native
from .data_type import BasicType, DataType

if typing.TYPE_CHECKING:
    from ..common import Context
    from .constrained import Constrained
    from .custom import Custom
    from .structure import Structure

T = typing.TypeVar("T", bound=list[Native])


@dataclasses.dataclass
class List(DataType[T]):
    """
    An ordered list with entries of the same SiLA data type.

    Attributes:
      data_type: The SiLA data type of each list item.
      value: The list of SiLA data type instances.
    """

    data_type: typing.ClassVar[type[typing.Union[BasicType, "Structure", "Constrained", "Custom"]]] = Any

    value: collections.abc.Sequence[typing.Union[BasicType, "Structure", "Constrained", "Custom"]] = dataclasses.field(
        default_factory=list
    )

    @typing.override
    @classmethod
    async def from_native(cls, context: "Context", value: typing.Optional[T] = None, /) -> typing.Self:
        value = value or []

        values = []
        for index, item in enumerate(value):
            try:
                values.append(await cls.data_type.from_native(context, item))
            except ConversionError as error:
                raise ConversionError(error.message, [str(index), *error.path]) from error
            except Exception as error:  # noqa: BLE001
                raise ConversionError(str(error), [str(index)]) from None

        return await cls(value=values).validate()

    @typing.override
    async def to_native(self, context: "Context", /) -> T:
        await self.validate()

        values: T = typing.cast(T, [])
        for index, item in enumerate(self.value):
            if not isinstance(item, self.data_type):
                msg = f"Expected value of type '{self.data_type.__name__}', received '{type(item).__name__}'."
                raise ConversionError(msg, [str(index)])

            try:
                values.append(await item.to_native(context))
            except ConversionError as error:
                raise ConversionError(error.message, [str(index), *error.path]) from error
            except Exception as error:  # noqa: BLE001
                raise ConversionError(str(error), [str(index)]) from None

        return values

    @typing.override
    async def validate(self) -> typing.Self:
        return self

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = cls()
        end = reader.length if length is None else reader.cursor + length

        values = []
        field_number = 0
        while reader.cursor < end:
            tag = reader.read_uint32()

            if field_number == 0:
                field_number = tag >> 3
            elif field_number != tag >> 3:
                reader.skip_type(tag & 7)
                continue

            try:
                reader.expect_type(tag, WireType.LEN)
                value = message.data_type.decode(reader, reader.read_uint32())
                values.append(value)
            except DecodeError as error:
                raise DecodeError(error.message, error.offset, [len(values), *error.path]) from None

        message.value = values

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        for item in self.value:
            item.encode(writer, number or 1)

        return writer.finish()

    @typing.override
    @classmethod
    def serialize(cls, serializer: Serializer) -> None:
        serializer.start_element("DataType")
        serializer.start_element("List")
        cls.data_type.serialize(serializer)
        serializer.end_element("List")
        serializer.end_element("DataType")

    @typing.override
    @classmethod
    def deserialize(cls, deserializer: Deserializer) -> collections.abc.Generator[None, typing.Any, type[typing.Self]]:
        from .constrained import Constrained
        from .custom import Custom
        from .structure import Structure

        yield from deserializer.read_start_element(name="List")

        data_type = yield from deserializer.read(DataType.deserialize(deserializer))

        if issubclass(data_type, cls):
            msg = "The data type of list entries must not be a list itself."
            raise ValueError(msg)

        assert (
            issubclass(data_type, BasicType)
            or issubclass(data_type, Structure)
            or issubclass(data_type, Constrained)
            or issubclass(data_type, Custom)
        )

        yield from deserializer.read_end_element(name="List")

        return cls.create(data_type)

    @classmethod
    def create(
        cls,
        data_type: type[typing.Union[BasicType, "Structure", "Constrained", "Custom"]] = Any,
        name: str = "",
    ) -> type[typing.Self]:
        """
        Create a new SiLA `List` class with the provided data type.

        Args:
          data_type: The SiLA data type of each list item.
          name: An optional name for the new `List` class.

        Returns:
          A new `List` class with the specified data type.
        """

        if issubclass(data_type, cls):
            msg = "The data type of list entries must not be a list itself."
            raise TypeError(msg)

        return dataclasses.make_dataclass(
            name or cls.__name__,
            [("value", collections.abc.Sequence[data_type], dataclasses.field(default_factory=list))],
            bases=(cls,),
            namespace={"data_type": data_type},
            eq=False,
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, List):
            return NotImplemented

        return (
            self.data_type.__name__ == other.data_type.__name__
            and len(self.value) == len(other.value)
            and all(a == b for a, b in zip(self.value, other.value))
        )



import collections.abc
import datetime

import typing_extensions as typing

if typing.TYPE_CHECKING:
    from ..common import Context


Native = typing.Union[
    str,
    int,
    float,
    bool,
    bytes,
    datetime.date,
    datetime.time,
    datetime.datetime,
    collections.abc.Sequence["Native"],
    collections.abc.Mapping[str, "Native"],
]

T = typing.TypeVar("T", bound=Native)


class Convertable(typing.Protocol[T]):
    """Convert a native Python value to this class and back."""

    @classmethod
    async def from_native(cls, context: "Context", value: typing.Optional[T] = None, /) -> typing.Self:
        """
        Convert a native Python value to its corresponding SiLA data type.

        Args:
          context: The context in which the conversion is performed.
          value: The native value to convert. If not provided, an
            instance with a default value is created.

        Returns:
          An instance of the corresponding SiLA data type.

        Raises:
          ConversionError: If the provided native value cannot be
            converted to the SiLA data type.
        """
        ...

    async def to_native(self, context: "Context", /) -> T:
        """
        Convert the SiLA data type to a native Python value.

        Args:
          context: The context in which the conversion is performed.

        Returns:
          The native Python value corresponding to the SiLA data type.

        Raises:
          ConversionError: If the SiLA data type cannot be converted to a
            native Python value.
        """
        ...

    async def validate(self) -> typing.Self:
        """
        Test constraints and restrictions of the data type value.

        Returns:
          The data type instance, allowing for method chaining.

        Raises:
          TypeError: If the value has the incorrect value.
          ValueError: If the value does not comply with the given
            constraints and restrictions.
        """
        ...



import collections.abc
import dataclasses

import typing_extensions as typing

from ..fdl import Characters, Deserializer, EndElement, Serializer, StartElement
from ..protobuf import ConversionError, DecodeError, Reader, WireType, Writer
from ..validators import check_display_name, check_identifier
from .any import Any
from .convertable import Native
from .data_type import DataType
from .element import Element

if typing.TYPE_CHECKING:
    from ..common import Context

T = typing.TypeVar("T", bound=dict[str, Native])


@dataclasses.dataclass
class Structure(DataType[T]):
    """
    A structure composed of one or more named elements with the same or different SiLA types.

    Attributes:
      presence: A list of element names that were present in the
        protobuf message.
      **kwargs: Key and value pairs of the SiLA structure instances.
    """

    elements: typing.ClassVar[dict[str, Element]] = {}

    value: dict[str, DataType] = dataclasses.field(default_factory=dict)

    @typing.override
    @classmethod
    async def from_native(cls, context: "Context", value: typing.Optional[T] = None, /) -> typing.Self:
        value = value or {}

        structure = cls()
        for name, element in cls.elements.items():
            item = value.get(name, None)

            if item is None and not issubclass(element.data_type, Any):
                msg = f"Missing field '{element.identifier}' in message '{cls.__name__}'."
                raise ConversionError(msg, [element.identifier])

            try:
                structure.value[name] = await element.data_type.from_native(context, item)
            except ConversionError as error:
                raise ConversionError(error.message, [element.identifier, *error.path]) from error
            except Exception as error:  # noqa: BLE001
                raise ConversionError(str(error), [element.identifier]) from None

        return await structure.validate()

    @typing.override
    async def to_native(self, context: "Context", /) -> T:
        await self.validate()

        values: T = typing.cast(T, {})
        for name, element in self.elements.items():
            if name not in self.value:
                msg = f"Missing field '{element.identifier}' in message '{self.__class__.__name__}'."
                raise ConversionError(msg, [element.identifier])

            item = self.value.get(name, element.data_type())
            if not isinstance(item, element.data_type):
                msg = f"Expected value of type '{element.data_type.__name__}', received '{type(item).__name__}'."
                raise ConversionError(msg, [element.identifier])

            try:
                values[name] = await item.to_native(context)
            except ConversionError as error:
                raise ConversionError(error.message, [element.identifier, *error.path]) from None
            except Exception as error:  # noqa: BLE001
                raise ConversionError(str(error), [element.identifier]) from None

        return values

    @typing.override
    async def validate(self) -> typing.Self:
        return self

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        from .constrained import Constrained
        from .list import List

        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = cls()
        end = reader.length if length is None else reader.cursor + length

        elements = list(cls.elements.items())
        while reader.cursor < end:
            pos = reader.cursor
            tag = reader.read_uint32()
            index = (tag >> 3) - 1

            if index >= len(elements):
                reader.skip_type(tag & 7)
                continue

            reader.expect_type(tag, WireType.LEN)
            name, element = elements[index]

            try:
                if issubclass(element.data_type, List) or (
                    issubclass(element.data_type, Constrained) and issubclass(element.data_type.data_type, List)
                ):
                    if name not in message.value:
                        values = element.data_type.decode(reader.buffer[pos:end])
                        message.value[name] = values

                    reader.skip_type(tag & 7)
                else:
                    value = element.data_type.decode(reader, reader.read_uint32())
                    message.value[name] = value
            except DecodeError as error:
                msg = f"Invalid field '{element.identifier}' in message '{cls.__name__}': {error.message}"
                raise DecodeError(msg, error.offset, [element.identifier, *error.path]) from None

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        for field_number, (name, element) in enumerate(self.elements.items(), start=1):
            value = self.value.get(name, None) or element.data_type()

            value.encode(writer, field_number)

        if number:
            writer.ldelim()

        return writer.finish()

    @typing.override
    @classmethod
    def serialize(cls, serializer: Serializer) -> None:
        serializer.start_element("DataType")
        serializer.start_element("Structure")

        for element in cls.elements.values():
            serializer.start_element("Element")
            serializer.write_str("Identifier", element.identifier)
            serializer.write_str("DisplayName", element.display_name)
            serializer.write_str("Description", element.description)
            element.data_type.serialize(serializer)
            serializer.end_element("Element")

        serializer.end_element("Structure")
        serializer.end_element("DataType")

    @typing.override
    @classmethod
    def deserialize(cls, deserializer: Deserializer) -> collections.abc.Generator[None, typing.Any, type[typing.Self]]:
        yield from deserializer.read_start_element(name="Structure")

        elements: dict[str, Element] = {}
        while True:
            token = yield

            if isinstance(token, StartElement):
                if token.name == "Element":
                    # Identifier
                    yield from deserializer.read_start_element("Identifier")
                    identifier = yield from deserializer.read_str()
                    check_identifier(identifier)
                    yield from deserializer.read_end_element("Identifier")

                    # DisplayName
                    yield from deserializer.read_start_element("DisplayName")
                    display_name = yield from deserializer.read_str()
                    check_display_name(display_name)
                    yield from deserializer.read_end_element("DisplayName")

                    # Description
                    yield from deserializer.read_start_element("Description")
                    description = yield from deserializer.read_str()
                    yield from deserializer.read_end_element("Description")

                    # DataType
                    data_type = yield from deserializer.read(DataType.deserialize(deserializer))

                    elements[identifier] = Element(
                        identifier=identifier,
                        display_name=display_name,
                        description=description,
                        data_type=data_type,
                    )
                else:
                    msg = (
                        f"Expected start element with name 'Element', received start element with name '{token.name}'."
                    )
                    raise ValueError(msg)

            elif isinstance(token, EndElement):
                if token.name == "Element":
                    continue
                else:
                    break  # pragma: no cover

            elif isinstance(token, Characters):
                msg = f"Expected start element with name 'Element', received characters '{token.value}'."
                raise ValueError(msg)

        if not elements:
            msg = "Expected at least one 'Element' element inside the 'Structure' element."
            raise ValueError(msg)

        return cls.create(elements)

    @classmethod
    def create(
        cls,
        elements: typing.Optional[dict[str, Element]] = None,
        name: typing.Optional[str] = None,
        description: typing.Optional[str] = None,
    ) -> type[typing.Self]:
        """
        Create a new SiLA `Structure` class with the provided elements.

        Args:
          elements: A mapping of names and their corresponding SiLA
            elements.
          name: An optional name for the new `Structure` class.
          description: An optional description for the new `Structure` class.

        Returns:
          A new `Structure` class with the specified elements.
        """

        elements = elements or {}

        return dataclasses.make_dataclass(
            name or cls.__name__,
            [],
            bases=(cls,),
            namespace={"__doc__": description, "elements": elements},
            eq=False,
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Structure):
            return NotImplemented

        return (
            self.elements.keys() == other.elements.keys()
            and all(
                a.identifier == b.identifier
                and a.display_name == b.display_name
                and a.description == b.description
                and a.data_type.__name__ == b.data_type.__name__
                for a, b in zip(self.elements.values(), other.elements.values())
            )
            and self.value == other.value
        )



import collections.abc
import dataclasses
import functools

import typing_extensions as typing

from ..common import Handler
from ..fdl import Serializable
from ..identifiers import CommandIdentifier

if typing.TYPE_CHECKING:
    from ..common import Feature
    from ..data_types import Element
    from ..fdl import Deserializer, Serializer


@dataclasses.dataclass
class Command(Handler, Serializable):
    """Describes certain actions that can be performed on the server."""

    observable: bool = False
    """Whether the command execution is observable or not."""

    parameters: dict[str, "Element"] = dataclasses.field(default_factory=dict)
    """The parameters of the command."""

    responses: dict[str, "Element"] = dataclasses.field(default_factory=dict)
    """The responses of the command containing the result."""

    @functools.cached_property
    @typing.override
    def fully_qualified_identifier(self) -> CommandIdentifier:
        """Uniquely identifies the command."""

        return CommandIdentifier.create(**super().fully_qualified_identifier._data, command=self.identifier)

    @typing.override
    def add_to_feature(self, feature: "Feature") -> typing.Self:
        super().add_to_feature(feature)

        feature.commands[self.identifier] = self

        return self

    @typing.override
    def serialize(self, serializer: "Serializer") -> None:
        from .observable_command import ObservableCommand

        serializer.start_element("Command")
        serializer.write_str("Identifier", self.identifier)
        serializer.write_str("DisplayName", self.display_name)
        serializer.write_str("Description", self.description)
        serializer.write_bool("Observable", self.observable)

        for element in self.parameters.values():
            serializer.start_element("Parameter")
            serializer.write_str("Identifier", element.identifier)
            serializer.write_str("DisplayName", element.display_name)
            serializer.write_str("Description", element.description)
            element.data_type.serialize(serializer)
            serializer.end_element("Parameter")

        for element in self.responses.values():
            serializer.start_element("Response")
            serializer.write_str("Identifier", element.identifier)
            serializer.write_str("DisplayName", element.display_name)
            serializer.write_str("Description", element.description)
            element.data_type.serialize(serializer)
            serializer.end_element("Response")

        if isinstance(self, ObservableCommand):
            for element in self.intermediate_responses.values():
                serializer.start_element("IntermediateResponse")
                serializer.write_str("Identifier", element.identifier)
                serializer.write_str("DisplayName", element.display_name)
                serializer.write_str("Description", element.description)
                element.data_type.serialize(serializer)
                serializer.end_element("IntermediateResponse")

        if self.errors:
            serializer.start_element("DefinedExecutionErrors")
            for Error in self.errors.values():
                serializer.write_str("Identifier", Error.identifier)
            serializer.end_element("DefinedExecutionErrors")

        serializer.end_element("Command")

    @typing.override
    def deserialize(self, deserializer: "Deserializer") -> collections.abc.Generator[None, typing.Any, typing.Self]: ...



import abc
import asyncio
import dataclasses
import logging

import typing_extensions as typing

from ..identifiers import (
    CommandIdentifier,
    DataTypeIdentifier,
    FeatureIdentifier,
    MetadataIdentifier,
    PropertyIdentifier,
)
from .context import Context

if typing.TYPE_CHECKING:
    from ...server import (
        CommandExecution,
        Metadata,
        ObservableCommand,
        ObservableProperty,
        ServerBinaryTransferHandler,
        UnobservableCommand,
        UnobservableProperty,
    )
    from ..command import Command
    from ..data_types import Custom
    from ..property import Property
    from .feature import Feature


@dataclasses.dataclass
class Server(Context):
    """A system (a software system, a laboratory instrument, or device) that offers features to a client."""

    def __init__(self) -> None:
        super().__init__()

        from ...server import ServerBinaryTransferHandler

        self._ready = asyncio.Event()
        self._shutdown = asyncio.Event()
        self._command_executions: dict[str, "CommandExecution"] = {}
        self._binary_transfer_handler: "ServerBinaryTransferHandler" = ServerBinaryTransferHandler(self)

    @property
    def logger(self) -> logging.Logger:
        """A python logger instance."""

        return logging.getLogger(__name__)

    @abc.abstractmethod
    async def start(self) -> None:
        """Start this server."""

        if self._ready.is_set():
            msg = f"The {self.__class__.__name__} is already running."
            raise RuntimeError(msg)

        self.logger.info(f"Starting {self.__class__.__name__}...")
        self._shutdown.clear()
        await self._start()
        self._ready.set()

    @abc.abstractmethod
    async def _start(self) -> None:
        """Override this method to implement the start behaviour of the server."""

    @abc.abstractmethod
    async def stop(self, grace: typing.Optional[float] = None) -> None:
        """
        Stop this server.

        Args:
          grace: A grace period in seconds allowing the RPC handlers to
            gracefully shutdown.
        """

        self.logger.info(f"Stopping {self.__class__.__name__}...")
        self._ready.clear()
        self._shutdown.set()
        await self._stop(grace)

    @abc.abstractmethod
    async def _stop(self, grace: typing.Optional[float] = None) -> None:
        """
        Override this method to implement the stop behaviour of the server.

        Args:
          grace: A grace period in seconds allowing the RPC handlers to
            gracefully shutdown.
        """

    async def wait_for_ready(self) -> None:
        """Wait until the server is ready."""

        await self._ready.wait()
        self._ready.clear()

    async def wait_for_termination(self) -> None:
        """Continues current coroutine once the server stops."""

        await self._shutdown.wait()
        self._shutdown.clear()

    @typing.override
    def get_command_execution(self, command_execution_uuid: str) -> "CommandExecution":
        return typing.cast("CommandExecution", super().get_command_execution(command_execution_uuid))

    @typing.override
    def register_feature(self, feature: "Feature") -> None:
        """
        Add a feature as RPC handler with this server.

        Args:
          feature: The SiLA feature to add to this server.
        """

        if self._ready.is_set():
            msg = f"Unable to register feature. The {self.__class__.__name__} is already running."
            raise RuntimeError(msg)

        super().register_feature(feature)

    def get_feature(self, identifier: str) -> "Feature":
        """
        Get a registered feature by its identifier.

        Args:
          identifier: The fully qualified identifier of the feature to
            receive.

        Returns:
          The feature registered with this server.

        Raises:
          ValueError: If the given identifier is invalid or not
            recognized.
        """

        identifier = FeatureIdentifier(identifier)

        if identifier.feature_identifier not in self.features:
            msg = f"Requested unknown feature identifier '{identifier.feature_identifier}'."
            raise ValueError(msg)

        return self.features[identifier.feature_identifier]

    def get_property(self, identifier: str) -> "Property":
        """
        Get a property by its identifier.

        Args:
          identifier: The fully qualified identifier of the property to
            receive.

        Returns:
          The property registered with this server.

        Raises:
          ValueError: If the given identifier is invalid or not
            recognized.
        """

        identifier = PropertyIdentifier(identifier)

        feature = self.get_feature(identifier)

        if identifier.property not in feature.properties:
            msg = f"Requested unknown property identifier '{identifier}'."
            raise ValueError(msg)

        return feature.properties[identifier.property]

    def get_unobservable_property(self, identifier: str) -> "UnobservableProperty":
        """
        Get an unobservable property by its identifier.

        Args:
          identifier: The fully qualified identifier of the unobservable
            property to receive.

        Returns:
          The unobservable property registered with this server.

        Raises:
          ValueError: If the given identifier is invalid or not
            recognized.
        """

        property_ = self.get_property(identifier)

        from ...server import UnobservableProperty

        if not isinstance(property_, UnobservableProperty):
            msg = "Expected identifier to reference an unobservable property, received 'observable property' instead."
            raise ValueError(msg)

        return property_

    def get_observable_property(self, identifier: str) -> "ObservableProperty":
        """
        Get an observable property by its identifier.

        Args:
          identifier: The fully qualified identifier of the observable
            property to receive.

        Returns:
          The observable property registered with this server.

        Raises:
          ValueError: If the given identifier is invalid or not
            recognized.
        """

        property_ = self.get_property(identifier)

        from ...server import ObservableProperty

        if not isinstance(property_, ObservableProperty):
            msg = "Expected identifier to reference an observable property, received 'unobservable property' instead."
            raise ValueError(msg)

        return property_

    def get_command(self, identifier: str) -> "Command":
        """
        Get a command by its identifier.

        Args:
          identifier: The fully qualified identifier of the command to
            receive.

        Returns:
          The command registered with this server.

        Raises:
          ValueError: If the given identifier is invalid or not
            recognized.
        """

        identifier = CommandIdentifier(identifier)

        feature = self.get_feature(identifier)

        if identifier.command not in feature.commands:
            msg = f"Requested unknown command identifier '{identifier}'."
            raise ValueError(msg)

        return feature.commands[identifier.command]

    def get_unobservable_command(self, identifier: str) -> "UnobservableCommand":
        """
        Get an unobservable command by its identifier.

        Args:
          identifier: The fully qualified identifier of the unobservable
            command to receive.

        Returns:
          The unobservable command registered with this server.

        Raises:
          ValueError: If the given identifier is invalid or not
            recognized.
        """

        command = self.get_command(identifier)

        from ...server import UnobservableCommand

        if not isinstance(command, UnobservableCommand):
            msg = "Expected identifier to reference an unobservable command, received 'observable command' instead."
            raise ValueError(msg)

        return command

    def get_observable_command(self, identifier: str) -> "ObservableCommand":
        """
        Get an observable command by its identifier.

        Args:
          identifier: The fully qualified identifier of the observable
            command to receive.

        Returns:
          The observable command registered with this server.

        Raises:
          ValueError: If the given identifier is invalid or not
            recognized.
        """

        command = self.get_command(identifier)

        from ...server import ObservableCommand

        if not isinstance(command, ObservableCommand):
            msg = "Expected identifier to reference an observable command, received 'unobservable command' instead."
            raise ValueError(msg)

        return command

    def get_data_type_definition(self, identifier: str) -> type["Custom"]:
        """
        Get a custom data type definition by its identifier.

        Args:
          identifier: The fully qualified identifier of the custom data
            type definition to receive.

        Returns:
          The custom data type definition registered with this server.

        Raises:
          ValueError: If the given identifier is invalid or not
            recognized.
        """

        identifier = DataTypeIdentifier(identifier)

        feature = self.get_feature(identifier)

        if identifier.data_type not in feature.data_type_definitions:
            msg = f"Requested unknown custom data type identifier '{identifier}'."
            raise ValueError(msg)

        return feature.data_type_definitions[identifier.data_type]

    def get_metadata(self, identifier: str) -> type["Metadata"]:
        """
        Get metadata by its identifier.

        Args:
          identifier: The fully qualified identifier of the metadata to
            receive.

        Returns:
          The metadata registered with this server.

        Raises:
          ValueError: If the given identifier is invalid or not
            recognized.
        """

        identifier = MetadataIdentifier(identifier)

        feature = self.get_feature(identifier)

        if identifier.metadata not in feature.metadata:
            msg = f"Requested unknown metadata identifier '{identifier}'."
            raise ValueError(msg)

        from ...server import Metadata

        metadata = feature.metadata[identifier.metadata]
        assert issubclass(metadata, Metadata)

        return metadata

    def get_metadata_by_affect(self, identifier: str) -> list[type["Metadata"]]:
        """
        Get a list of metadata that affect the given identifier.

        Args:
          identifier: The fully qualified identifier of the feature,
            command or property to check.

        Returns:
          The metadata that affects the given identifier.

        Raises:
          ValueError: If the given identifier is invalid or not
            recognized.
        """

        identifier = FeatureIdentifier(identifier)

        from ...server import Metadata

        return [
            metadata
            for feature in self.features.values()
            for metadata in feature.metadata.values()
            if issubclass(metadata, Metadata)
            and (identifier.feature_identifier in metadata.affects or identifier in metadata.affects)
        ]



import dataclasses
import functools
import logging
import uuid

import grpc.aio
import typing_extensions as typing

from sila import framework

from ..framework import ChannelOptions, VersionLevel, check_display_name, check_identifier, check_uuid, check_version
from ..framework.grpc import create_server


@dataclasses.dataclass
class ServerConfig:
    """Configuration to run a SiLA 2 server."""

    hostname: str = "0.0.0.0"
    """The target hostname to bind to. Defaults to `0.0.0.0`."""

    port: int = 0
    """
    The target port to bind to. If set to `0` an available port is
    chosen at runtime. Defaults to `0`.
    """

    tls: bool = False
    """
    Whether to run a secure/TLS server or a plaintext server (i.e. no
    TLS), defaults to run with TLS encryption.
    """

    require_client_auth: bool = False
    """
    A boolean indicating whether or not to require clients to be
    authenticated. May only be True if root_certificates is not None.
    """

    root_certificates: typing.Optional[bytes] = None
    """
    The PEM-encoded root certificates as a byte string, or None to
    retrieve them from a default location chosen by gRPC runtime.
    """

    certificate_chain: typing.Optional[bytes] = None
    """
    The PEM-encoded certificate chain as a byte string to use or None
    if no certificate chain should be used.
    """

    private_key: typing.Optional[bytes] = None
    """
    The PEM-encoded private key as a byte string, or None if no
    private key should be used.
    """

    options: ChannelOptions = dataclasses.field(default_factory=ChannelOptions)
    """
    Additional options for the underlying gRPC connection.
    """

    uuid: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))
    """
    Uniquely identifies the SiLA server. Needs to remain the same even after restarting the server.
    Follows the textual representation of UUIDs, e.g. "082bc5dc-18ae-4e17-b028-6115bbc6d21e".
    """

    name: str = "SiLA Server"
    """
    Human readable name of the SiLA server. This value is configurable during runtime via the SiLA
    Service feature's `set_server_name` command. Must not exceed 255 characters.
    """

    type: str = "ExampleServer"
    """
    Human readable identifier of the SiLA server used to describe the entity the server represents.
    Starts with a capital letter, continued by letters and digits up to a maximum of 255 characters.
    """

    description: str = ""
    """Describes the use and purpose of the SiLA Server."""

    version: str = "0.1.0"
    """
    The version of the SiLA server following the Semantic Versioning specification with pre-release
    identifiers separated by underscores, e.g. "3.19.373_mighty_lab_devices".
    """

    vendor_url: str = "https://sila-standard.com"
    """
    URL to the website of the vendor or the website of the product of this SiLA Server. Follows the
    Uniform Resource Locator specification in RFC 1738.
    """


class Server(framework.Server):
    """
    A SiLA 2 compliant gRPC server.

    A SiLA Server can either be a physical laboratory instrument or a software system that offers
    functionalities to clients. These functions are specified and described in Features.
    """

    uuid: str = ""
    """
    The SiLA server UUID is a UUID of a SiLA server. Each SiLA server must generate a UUID once, to uniquely identify
    itself. It needs to remain the same even after the lifetime of a SiLA server has ended.
    """

    type: str = ""
    """
    The SiLA server type is a human readable identifier of the SiLA server used to describe the entity that the SiLA
    server represents. For example, the make and model for a hardware device. A SiLA server type must comply with the
    rules for any identifier and start with an upper-case letter (A-Z) and may be continued by lower and upper-case
    letters (A-Z and a-z) and digits (0-9) up to a maximum of 255 characters in length.
    """

    name: str = ""
    """
    The SiLA server name is a human readable name of the SiLA server. By default this name should be equal to the SiLA
    server type. This property must be configurable via the SiLA service feature's “Set Server Name” command. This
    property has no uniqueness guarantee. A SiLA server name is the display name of a SiLA server (i.e. must comply with
    the rules for any display name, hence be a string of unicode characters of maximum 255 characters in length).
    """

    version: str = "0.1.0"
    """
    The SiLA server version is the version of the SiLA server. A "Major" and a "Minor" version number (e.g. 1.0) must be
    provided, a "Patch" version number may be provided. Optionally, an arbitrary text, separated by an underscore may be
    appended, e.g. “3.19.373_mighty_lab_devices”.
    """

    description: str = ""
    """
    The SiLA server description is the description of the SiLA server. It should include the use and purpose of this
    SiLA server.
    """

    vendor_url: str = ""
    """
    The SiLA server vendor URL is the URL to the website of the vendor or the website of the product of this SiLA
    server. This URL should be accessible at all times. The URL is a Uniform Resource Locator as defined in RFC 1738.
    """

    def __init__(self, config: typing.Optional[typing.Union[dict, ServerConfig]] = None, /):
        super().__init__()

        if isinstance(config, dict):
            import warnings

            warnings.warn(
                "Providing ServerConfig as a dictionary is deprecated and will be removed in a future release. "
                "Please provide a ServerConfig instance instead.",
                category=DeprecationWarning,
            )
            self._config = ServerConfig(**config)
        else:
            self._config = config or ServerConfig()

        self._address = f"{self._config.hostname}:{self._config.port}"
        self._options = self._config.options
        self._server: typing.Optional[grpc.aio.Server] = None

        self.uuid = check_uuid(self._config.uuid)
        self.type = check_identifier(self._config.type)
        self.name = check_display_name(self._config.name)
        self.version = check_version(self._config.version, required=VersionLevel.MINOR, optional=VersionLevel.LABEL)
        self.description = self._config.description
        self.vendor_url = self._config.vendor_url

        from .sila_service import SiLAService

        self.register_feature(SiLAService())

    @property
    def logger(self) -> logging.Logger:
        """A python logger instance."""

        return logging.getLogger(__name__)

    @typing.override
    async def _start(self) -> None:
        await super()._start()

        self._server, port = await functools.partial(create_server, **dataclasses.asdict(self._config))(self._address)
        self._address = f"{self._address.rpartition(':')[0]}:{port}"
        self.logger.info(f"{self.__class__.__name__} bound to '{self._address}'.")

        for service, method_handlers in self.protobuf.services.items():
            handlers = grpc.method_handlers_generic_handler(service, method_handlers)
            self._server.add_generic_rpc_handlers((handlers,))

        await self._server.start()

    @typing.override
    async def _stop(self, grace: typing.Optional[float] = None) -> None:
        await super()._stop(grace)

        if self._server is not None:
            await self._server.stop(grace=grace)



from .conversion_error import ConversionError
from .decode_error import DecodeError
from .encode_error import EncodeError
from .message import Message
from .protobuf import Protobuf
from .reader import Reader
from .wire_type import WireType
from .writer import Writer

__all__ = [
    "ConversionError",
    "DecodeError",
    "EncodeError",
    "Message",
    "Protobuf",
    "Reader",
    "WireType",
    "Writer",
]



import collections.abc
import uuid
import weakref

import grpc
import grpc.aio
import typing_extensions as typing

from ..framework.binary_transfer import (
    BinaryDownloadFailed,
    BinaryTransfer,
    BinaryTransferError,
    BinaryTransferHandler,
    BinaryUploadFailed,
    CreateBinaryRequest,
    CreateBinaryResponse,
    DeleteBinaryRequest,
    DeleteBinaryResponse,
    DownloadChunkRequest,
    DownloadChunkResponse,
    GetBinaryInfoRequest,
    GetBinaryInfoResponse,
    InvalidBinaryTransferUUID,
    UploadChunkRequest,
    UploadChunkResponse,
)
from ..framework.data_types import Binary, Constrained, Custom, DataType, Duration, List, Structure
from ..framework.errors import SiLAError
from ..framework.identifiers import ParameterIdentifier

if typing.TYPE_CHECKING:
    from ..framework.common import Server


class ServerBinaryTransferHandler(BinaryTransferHandler):
    """
    Handle the transfer of large binaries on the server side.

    Args:
      context: The server this handler will be attached to.
    """

    def __init__(self, context: "Server"):
        self._binaries: dict[str, BinaryTransfer] = {}
        self._context: "Server" = weakref.proxy(context)

        self.add_to_server(self._context)

    @property
    def context(self) -> "Server":
        """The context in which the binary transfer handler is applied."""

        return self._context

    @typing.override
    async def get_binary(self, binary_transfer_uuid: str) -> bytes:
        binary = self.get_binary_by_uuid(binary_transfer_uuid)

        if not binary.is_complete:
            msg = f"Requested incomplete Binary with 'binary_transfer_uuid' of '{binary_transfer_uuid}'."
            raise ValueError(msg)

        return binary.buffer

    @typing.override
    async def set_binary(self, value: bytes) -> str:
        binary = BinaryTransfer.from_buffer(value)
        self._binaries[binary.binary_transfer_uuid] = binary

        return binary.binary_transfer_uuid

    def add_to_server(self, server: "Server") -> None:
        """
        Add the necessary grpc handlers to the given server.

        Args:
          server: The server to attach the handlers to.
        """

        server.protobuf.register_service(
            "BinaryUpload",
            {
                "CreateBinary": grpc.unary_unary_rpc_method_handler(
                    self._create_binary,
                    request_deserializer=lambda x: CreateBinaryRequest.decode(x),
                    response_serializer=lambda x: x.encode(),
                ),
                "UploadChunk": grpc.stream_stream_rpc_method_handler(
                    self._upload_chunk,
                    request_deserializer=lambda x: UploadChunkRequest.decode(x),
                    response_serializer=lambda x: x.encode(),
                ),
                "DeleteBinary": grpc.unary_unary_rpc_method_handler(
                    self._delete_binary,
                    request_deserializer=lambda x: DeleteBinaryRequest.decode(x),
                    response_serializer=lambda x: x.encode(),
                ),
            },
            package="sila2.org.silastandard",
        )
        server.protobuf.register_service(
            "BinaryDownload",
            {
                "GetBinaryInfo": grpc.unary_unary_rpc_method_handler(
                    self._get_binary_info,
                    request_deserializer=lambda x: GetBinaryInfoRequest.decode(x),
                    response_serializer=lambda x: x.encode(),
                ),
                "GetChunk": grpc.stream_stream_rpc_method_handler(
                    self._download_chunk,
                    request_deserializer=lambda x: DownloadChunkRequest.decode(x),
                    response_serializer=lambda x: x.encode(),
                ),
                "DeleteBinary": grpc.unary_unary_rpc_method_handler(
                    self._delete_binary,
                    request_deserializer=lambda x: DeleteBinaryRequest.decode(x),
                    response_serializer=lambda x: x.encode(),
                ),
            },
            package="sila2.org.silastandard",
        )

    async def create_binary(
        self, request: CreateBinaryRequest, metadata: typing.Optional[dict[str, bytes]] = None
    ) -> CreateBinaryResponse:
        """
        Create a new binary transfer session.

        Args:
          request: The client's request to create a new binary.
          metadata: Additional metadata sent from client to server.

        Returns:
          Information about the newly created binary.

        Raises:
          BinaryUploadFailed: If an error occurs during creation.
        """

        try:
            identifier = ParameterIdentifier(request.parameter_identifier)
        except ValueError:
            msg = f"Expected a valid fully qualified parameter identifier, received '{request.parameter_identifier}'."
            raise BinaryUploadFailed(msg) from None

        try:
            command = self.context.get_command(identifier)
            parameter = next(
                parameter for parameter in command.parameters.values() if parameter.identifier == identifier.parameter
            )
        except (KeyError, ValueError, StopIteration):
            msg = f"Expected a known fully qualified parameter identifier, received '{request.parameter_identifier}'."
            raise BinaryUploadFailed(msg) from None

        for interceptor in self.context.get_metadata_by_affect(identifier.command_identifier):
            metadatum = await interceptor.from_buffer(command, metadata)
            await metadatum.intercept(command)

        if not self.has_binary(parameter.data_type):
            msg = "Expected a fully qualified parameter identifier containing a 'Binary'."
            raise BinaryUploadFailed(msg)

        binary = BinaryTransfer.new(size=request.binary_size, chunks=request.chunk_count)
        self._binaries[binary.binary_transfer_uuid] = binary

        return CreateBinaryResponse(
            binary_transfer_uuid=binary.binary_transfer_uuid,
            lifetime_of_binary=Duration.from_total_seconds(binary.lifetime.total_seconds()),
        )

    async def upload_chunk(self, request: UploadChunkRequest) -> UploadChunkResponse:
        """
        Upload a chunk of data as part of a binary transfer session.

        Args:
          request: The client's requests to upload an individual chunk.

        Returns:
          Information to acknowledge the successful upload of the chunk.

        Raises:
          InvalidBinaryTransferUUID: If the given identifier is invalid
            or not recognized.
          BinaryUploadFailed: If an error occurs during the upload.
        """

        try:
            binary = self.get_binary_by_uuid(request.binary_transfer_uuid)
            binary.set_chunk(request.chunk_index, request.payload)
        except ValueError as error:
            raise BinaryUploadFailed(str(error)) from None

        return UploadChunkResponse(
            binary_transfer_uuid=request.binary_transfer_uuid,
            chunk_index=request.chunk_index,
            lifetime_of_binary=Duration.from_total_seconds(binary.lifetime.total_seconds()),
        )

    async def get_binary_info(self, request: GetBinaryInfoRequest) -> GetBinaryInfoResponse:
        """
        Retrieve information about a specific binary transfer session.

        Args:
          request: The client's request containing the binary transfer
            uuid for which to get the details.

        Returns:
          Information about the binary transfer.

        Raises:
          InvalidBinaryTransferUUID: If the given identifier is invalid
            or not recognized.
        """

        try:
            binary = self.get_binary_by_uuid(request.binary_transfer_uuid)
        except BinaryTransferError as error:
            raise error from None

        return GetBinaryInfoResponse(
            binary_size=binary.size,
            lifetime_of_binary=Duration.from_total_seconds(binary.lifetime.total_seconds()),
        )

    async def download_chunk(self, request: DownloadChunkRequest) -> DownloadChunkResponse:
        """
        Download a chunk of data as part of a binary transfer session.

        Args:
          request: The client's request to downlaod an individual chunk.

        Returns:
          The data of the individually requested chunk.

        Raises:
          InvalidBinaryTransferUUID: If the given identifier is invalid
            or not recognized.
          BinaryDownloadFailed: If an error occurs during the upload.
        """

        try:
            binary = self.get_binary_by_uuid(request.binary_transfer_uuid)
            payload = binary.get_chunk(request.offset, request.length)
        except ValueError as error:
            raise BinaryDownloadFailed(str(error)) from None
        except BinaryTransferError as error:
            raise error from None

        return DownloadChunkResponse(
            binary_transfer_uuid=request.binary_transfer_uuid,
            offset=request.offset,
            payload=payload,
            lifetime_of_binary=Duration.from_total_seconds(binary.lifetime.total_seconds()),
        )

    async def delete_binary(self, request: DeleteBinaryRequest) -> DeleteBinaryResponse:
        """
        Delete all data from a binary transfer session.

        Args:
          request: The client's request containing the binary transfer
            uuid for which to delete the details.

        Returns:
          Confirm the deletion of a binary transfer session.

        Raises:
          InvalidBinaryTransferUUID: If the given identifier is invalid
            or not recognized.
        """

        try:
            binary = self.get_binary_by_uuid(request.binary_transfer_uuid)
        except BinaryTransferError as error:
            raise error from None

        self._binaries.pop(binary.binary_transfer_uuid)
        return DeleteBinaryResponse()

    def get_binary_by_uuid(self, binary_transfer_uuid: str) -> BinaryTransfer:
        """
        Get a binary by its unique identifier (UUID).

        Args:
          binary_transfer_uuid: A unique identifier (UUID) for the binary
            transfer session from which to get the large binary.

        Returns:
          The binary transfer associated with the specified identifier.

        Raises:
          InvalidBinaryTransferUUID: If the given identifier is invalid
            or not recognized.
        """

        try:
            uuid.UUID(binary_transfer_uuid, version=4)
        except ValueError:
            msg = f"Expected 'binary_transfer_uuid' with format UUID, received '{binary_transfer_uuid}'."
            raise InvalidBinaryTransferUUID(msg) from None

        if binary_transfer_uuid not in self._binaries:
            msg = f"Requested unknown Binary with 'binary_transfer_uuid' of '{binary_transfer_uuid}'."
            raise InvalidBinaryTransferUUID(msg)

        binary = self._binaries[binary_transfer_uuid]

        if binary.lifetime.total_seconds() <= 0:
            del self._binaries[binary_transfer_uuid]
            msg = f"Requested Binary '{binary_transfer_uuid}' with exceeded lifetime."
            raise InvalidBinaryTransferUUID(msg)

        return binary

    @classmethod
    def has_binary(cls, data_type: type[DataType]) -> bool:
        """
        Check whether the given data type accepts binary data.

        Args:
          data_type: The data type to check.

        Returns:
          Whether the given data type contains a binary value.
        """

        if issubclass(data_type, Binary):
            return True
        if issubclass(data_type, (List, Constrained, Custom)):
            return cls.has_binary(data_type.data_type)
        if issubclass(data_type, Structure):
            return any(cls.has_binary(element.data_type) for element in data_type.elements.values())

        return False

    async def _create_binary(
        self, request: CreateBinaryRequest, context: grpc.aio.ServicerContext
    ) -> CreateBinaryResponse:
        """
        RPC handler to create a new binary transfer session.

        Args:
          request: The client's request to create a new binary.
          context: The gRPC call context.

        Returns:
          Information about the newly created binary.
        """

        try:
            metadata: dict[str, bytes] = {
                key: value
                for key, value in context.invocation_metadata() or ()
                if key.startswith("sila-") and key.endswith("-bin") and isinstance(value, bytes)
            }

            return await self.create_binary(request, metadata)
        except SiLAError as error:
            raise await error.to_rpc_error(context) from None

    async def _upload_chunk(
        self, request_iterator: collections.abc.AsyncIterator[UploadChunkRequest], context: grpc.aio.ServicerContext
    ) -> collections.abc.AsyncIterator[UploadChunkResponse]:
        """
        RPC handler to upload chunks of data as part of a binary transfer session.

        Args:
          request_iterator: The client's requests to upload individual
            chunks.
          context: The gRPC call context.

        Returns:
          Information to acknowledge the successful upload of a specific
          chunk.
        """

        try:
            async for request in request_iterator:
                yield await self.upload_chunk(request)
        except BinaryTransferError as error:
            raise await error.to_rpc_error(context) from None

    async def _get_binary_info(
        self, request: GetBinaryInfoRequest, context: grpc.aio.ServicerContext
    ) -> GetBinaryInfoResponse:
        """
        RPC handler to retrieve information about a specific binary transfer session.

        Args:
          request: The client's request containing the binary transfer
            uuid for which to get the details.
          context: The gRPC call context.

        Returns:
          Information about the binary transfer.
        """

        try:
            return await self.get_binary_info(request)
        except BinaryTransferError as error:
            raise await error.to_rpc_error(context) from None

    async def _download_chunk(
        self, request_iterator: collections.abc.AsyncIterator[DownloadChunkRequest], context: grpc.aio.ServicerContext
    ) -> collections.abc.AsyncIterator[DownloadChunkResponse]:
        """
        RPC handler to download chunks of data as part of a binary transfer session.

        Args:
          request_iterator: The client's requests to downlaod individual
            chunks.
          context: The gRPC call context.

        Returns:
          The data of the individually requested chunk.
        """

        try:
            async for request in request_iterator:
                yield await self.download_chunk(request)
        except BinaryTransferError as error:
            raise await error.to_rpc_error(context) from None

    async def _delete_binary(
        self, request: DeleteBinaryRequest, context: grpc.aio.ServicerContext
    ) -> DeleteBinaryResponse:
        """
        RPC handler to delete all data from a binary transfer session.

        Args:
          request: The client's request containing the binary transfer
            uuid for which to delete the details.
          context: The gRPC call context.

        Returns:
          Confirm the deletion of a binary transfer session.
        """

        try:
            return await self.delete_binary(request)
        except BinaryTransferError as error:
            raise await error.to_rpc_error(context) from None



from .binary_transfer import BinaryTransfer
from .binary_transfer_error import (
    BinaryDownloadFailed,
    BinaryTransferError,
    BinaryTransferErrorType,
    BinaryUploadFailed,
    InvalidBinaryTransferUUID,
)
from .binary_transfer_handler import BinaryTransferHandler
from .create_binary_request import CreateBinaryRequest
from .create_binary_response import CreateBinaryResponse
from .delete_binary_request import DeleteBinaryRequest
from .delete_binary_response import DeleteBinaryResponse
from .download_chunk_request import DownloadChunkRequest
from .download_chunk_response import DownloadChunkResponse
from .get_binary_info_request import GetBinaryInfoRequest
from .get_binary_info_response import GetBinaryInfoResponse
from .upload_chunk_request import UploadChunkRequest
from .upload_chunk_response import UploadChunkResponse

__all__ = [
    "BinaryDownloadFailed",
    "BinaryTransfer",
    "BinaryTransferError",
    "BinaryTransferErrorType",
    "BinaryTransferHandler",
    "BinaryUploadFailed",
    "CreateBinaryRequest",
    "CreateBinaryResponse",
    "DeleteBinaryRequest",
    "DeleteBinaryResponse",
    "DownloadChunkRequest",
    "DownloadChunkResponse",
    "GetBinaryInfoRequest",
    "GetBinaryInfoResponse",
    "InvalidBinaryTransferUUID",
    "UploadChunkRequest",
    "UploadChunkResponse",
]



def check_display_name(value: str) -> str:
    """
    Validate whether the given value is a human readable display name.

    Args:
      value: The input value to validate.

    Returns:
      The input value, allowing for method chaining.

    Raises:
      ValueError: If the provided value does not follow the expected
        schema.
    """

    if not isinstance(value, str):
        msg = f"Display name must be a string of unicode characters, received '{value}'."
        raise ValueError(msg)

    if not value:
        msg = f"Display name must not be empty, received '{value}'."
        raise ValueError(msg)

    if len(value) > 255:
        msg = f"Display name must not exceed 255 characters in length, received '{value}'."
        raise ValueError(msg)

    return value



import dataclasses

import typing_extensions as typing

from sila import datetime

from ..protobuf import Message, Reader, WireType, Writer
from .convertable import Convertable

if typing.TYPE_CHECKING:
    from ..common import Context


@dataclasses.dataclass
class Duration(Message, Convertable):
    """
    A signed, fixed-length span of time at nanosecond resolution.

    Attributes:
      seconds: The seconds of the duration value. Defaults to zero.
      nanos: The nanos of the duration value. Defaults to zero.
    """

    seconds: int = 0
    nanos: int = 0

    @property
    def total_seconds(self) -> float:
        """Total seconds in the duration."""

        return self.seconds + self.nanos / 1_000_000_000

    @classmethod
    def from_total_seconds(cls, total_seconds: float) -> typing.Self:
        """Create a duration from the given seconds."""

        seconds = int(total_seconds)
        nanos = round((total_seconds - seconds) * 1_000_000_000)

        # Handle overflow or underflow of nanos
        if nanos >= 1_000_000_000:
            seconds += 1
            nanos -= 1_000_000_000
        elif nanos <= -1_000_000_000:
            seconds -= 1
            nanos += 1_000_000_000

        return cls(seconds=seconds, nanos=nanos)

    @typing.overload
    @classmethod
    async def from_native(
        cls, context: "Context", value: typing.Optional[datetime.timedelta] = None, /
    ) -> typing.Self: ...

    @typing.overload
    @classmethod
    async def from_native(cls, context: "Context", /, *, seconds: int = 0, nanos: int = 0) -> typing.Self: ...

    @typing.override
    @classmethod
    async def from_native(
        cls,
        context: "Context",
        value: typing.Optional[datetime.timedelta] = None,
        /,
        *,
        seconds: int = 0,
        nanos: int = 0,
    ) -> typing.Self:
        if value is None:
            return await cls(seconds=seconds, nanos=nanos).validate()

        return await cls(seconds=value.days * 86400 + value.seconds, nanos=value.microseconds * 1000).validate()

    @typing.override
    async def to_native(self, context: "Context", /) -> datetime.timedelta:
        await self.validate()

        return datetime.timedelta(seconds=self.seconds, microseconds=int(self.nanos / 1000))

    @typing.override
    async def validate(self) -> typing.Self:
        if not isinstance(self.seconds, int):
            msg = f"Expected seconds of type 'int', received '{type(self.seconds).__name__}'."
            raise TypeError(msg)

        if not isinstance(self.nanos, int):
            msg = f"Expected nanos of type 'int', received '{type(self.nanos).__name__}'."
            raise TypeError(msg)

        if not (self.seconds >= 0):
            msg = f"Seconds must be a positive number, received '{self.seconds}'."
            raise ValueError(msg)

        if not (0 <= self.nanos < 1e9):
            msg = f"Nanos must be between 0 and 1e+9, received '{self.nanos}'."
            raise ValueError(msg)

        return self

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = cls()
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.VARINT)
                message.seconds = reader.read_int64()
            elif field_number == 2:
                reader.expect_type(tag, WireType.VARINT)
                message.nanos = reader.read_int32()
            else:
                reader.skip_type(tag & 7)

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        if self.seconds:
            writer.write_uint32(8).write_int64(self.seconds)
        if self.nanos:
            writer.write_uint32(16).write_int32(self.nanos)

        if number:
            writer.ldelim()

        return writer.finish()



from .context import Context
from .context_proxy import ContextProxy
from .feature import Feature
from .handler import Handler
from .server import Server
from .stream import Stream

__all__ = [
    "Context",
    "ContextProxy",
    "Feature",
    "Handler",
    "Server",
    "Stream",
]



import collections.abc
import dataclasses

import typing_extensions as typing

from ..data_types.binary import Binary
from ..data_types.string import String
from ..fdl import Deserializer, Serializer
from .constraint import Constraint

if typing.TYPE_CHECKING:
    from ..data_types import BasicType, List

T = typing.TypeVar("T", str, bytes)


@dataclasses.dataclass
class Length(Constraint[T]):
    """
    A constraint that enforces an exact length for a value of type
    `str` or `bytes`.

    This constraint ensures that a given string or bytes object
    has an exact length specified by the user. It raises errors
    if the input value's length does not match the expected length
    or if the input value is not of type `str` or `bytes`.

    Args:
      value: The exact length the value must have. It must be a
        non-negative integer and less than 2⁶³.

    Raises:
      ValueError: If `value` is negative or exceeds 2⁶³.
    """

    value: int
    """
    The exact length the value must have. It must be a
    non-negative integer and less than 2⁶³.
    """

    def __post_init__(self):
        if self.value < 0:
            raise ValueError(f"Length must be a non-negative integer, received '{self.value}'.")
        if self.value >= 2**63:
            raise ValueError(f"Length must be less than 2⁶³, received '{self.value}'.")

    async def validate(self, value: T) -> bool:
        if not isinstance(value, (str, bytes)):
            raise TypeError(f"Expected value of type 'str' or 'bytes', received '{type(value).__name__}'.")

        if len(value) != self.value:
            raise ValueError(f"Expected value with length '{self.value}', received '{value}'.")

        return True

    def serialize(self, serializer: Serializer) -> None:
        serializer.write_int("Length", self.value)

    @classmethod
    def deserialize(
        cls, deserializer: Deserializer, data_type: typing.Union[type["BasicType"], type["List"]]
    ) -> collections.abc.Generator[None, typing.Any, typing.Self]:
        if not issubclass(data_type, (String, Binary)):
            raise ValueError(
                f"Expected constraint's data type to be 'String' or 'Binary', received '{data_type.__name__}'."
            )

        yield from deserializer.read_start_element(name="Length")
        value = yield from deserializer.read_int()
        yield from deserializer.read_end_element(name="Length")

        return cls(value)



import collections.abc
import dataclasses

import typing_extensions as typing

from sila import datetime

from ..data_types.date import Date
from ..data_types.integer import Integer
from ..data_types.real import Real
from ..data_types.string import String
from ..data_types.time import Time
from ..data_types.timestamp import Timestamp
from ..fdl import Characters, Deserializer, EndElement, Serializer, StartElement
from .constraint import Constraint

if typing.TYPE_CHECKING:
    from ..data_types import BasicType, List

T = typing.TypeVar("T", str, int, float, datetime.date, datetime.time, datetime.datetime)


@dataclasses.dataclass
class Set(Constraint[T]):
    """
    A constraint that enforces that a value is part of a defined set
    of values.

    This class checks if a given value is one of the allowed values
    defined in the `values` attribute. The values can be of various
    types, including String, Integer, Real, Date, Time and Timestamp.

    Args:
      values: A sequence of allowed values. The value to be validated
        must be included in this sequence.

    Raises:
      ValueError: If the list of allowed values is empty.
    """

    values: collections.abc.Sequence[T]
    """
    A sequence of allowed values. The value to be validated
    must be included in this sequence.
    """

    def __post_init__(self):
        if not self.values:
            raise ValueError("The list of allowed values must not be empty.")

        self.__type = type(self.values[0])

        if any(type(value) is not self.__type for value in self.values):
            raise TypeError("The list of allowed values must all have the same type.")

    async def validate(self, value: T) -> bool:
        if not isinstance(value, self.__type):
            raise TypeError(f"Expected value of type '{self.__type.__name__}', received '{type(value).__name__}'.")

        if value not in self.values:
            raise ValueError(f"Value '{value}' is not in the set of allowed values.")

        return True

    def serialize(self, serializer: Serializer) -> None:
        serializer.start_element("Set")
        for value in self.values:
            if isinstance(value, str):
                serializer.write_str("Value", value)
            elif isinstance(value, int):
                serializer.write_int("Value", value)
            elif isinstance(value, float):
                serializer.write_float("Value", value)
            elif isinstance(value, datetime.date):
                serializer.write_date("Value", value)
            elif isinstance(value, datetime.time):
                serializer.write_time("Value", value)
            elif isinstance(value, datetime.datetime):
                serializer.write_datetime("Value", value)

        serializer.end_element("Set")

    @classmethod
    def deserialize(
        cls, deserializer: Deserializer, data_type: typing.Union[type["BasicType"], type["List"]]
    ) -> collections.abc.Generator[None, typing.Any, typing.Self]:
        if not issubclass(data_type, (String, Integer, Real, Date, Time, Timestamp)):
            raise ValueError(
                f"Expected constraint's data type to be 'String', 'Integer', 'Real', 'Date', 'Time' or 'Timestamp', received '{data_type.__name__}'."
            )

        yield from deserializer.read_start_element(name="Set")

        values: list[T] = []
        while True:
            token = yield

            if isinstance(token, StartElement):
                if token.name == "Value":
                    if issubclass(data_type, String):
                        value = yield from deserializer.read_str()
                    elif issubclass(data_type, Integer):
                        value = yield from deserializer.read_int()
                    elif issubclass(data_type, Real):
                        value = yield from deserializer.read_float()
                    elif issubclass(data_type, Date):
                        value = yield from deserializer.read_date()
                    elif issubclass(data_type, Time):
                        value = yield from deserializer.read_time()
                    elif issubclass(data_type, Timestamp):
                        value = yield from deserializer.read_datetime()

                    values.append(typing.cast(T, value))
                else:
                    raise ValueError(
                        f"Expected start element with name 'Value', received start element with name '{token.name}'."
                    )

            elif isinstance(token, EndElement):
                if token.name == "Value":
                    continue
                else:
                    break  # pragma: no cover

            elif isinstance(token, Characters):
                raise ValueError(f"Expected start element with name 'Value', received characters '{token.value}'.")

        if not values:
            raise ValueError("Expected at least one 'Value' element inside the 'Set' element.")

        return cls(values)



import dataclasses
import uuid

import typing_extensions as typing

from ..binary_transfer import (
    BinaryTransferError,
    CreateBinaryResponse,
    DeleteBinaryResponse,
    DownloadChunkResponse,
    GetBinaryInfoResponse,
    UploadChunkResponse,
)
from ..errors import SiLAError
from ..protobuf import Message, Reader, WireType, Writer
from .command_confirmation_response import CommandConfirmationResponse
from .command_execution_response import CommandExecutionResponse
from .metadata_response import MetadataResponse
from .observable_command_response import ObservableCommandResponse
from .property_response import PropertyResponse
from .unobservable_command_response import UnobservableCommandResponse


@dataclasses.dataclass
class ServerMessage(Message):
    """
    Message sent from server to client through the cloud stream.

    Attributes:
      request_uuid: Unique ID used to map responses to requests.
    """

    request_uuid: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))
    unobservable_command_response: typing.Optional[UnobservableCommandResponse] = None
    observable_command_confirmation: typing.Optional[CommandConfirmationResponse] = None
    observable_command_execution_info: typing.Optional[CommandExecutionResponse] = None
    observable_command_intermediate_response: typing.Optional[ObservableCommandResponse] = None
    observable_command_response: typing.Optional[ObservableCommandResponse] = None
    get_fcp_affected_by_metadata_response: typing.Optional[MetadataResponse] = None
    unobservable_property_value: typing.Optional[PropertyResponse] = None
    observable_property_value: typing.Optional[PropertyResponse] = None
    create_binary_response: typing.Optional[CreateBinaryResponse] = None
    upload_chunk_response: typing.Optional[UploadChunkResponse] = None
    delete_binary_response: typing.Optional[DeleteBinaryResponse] = None
    get_binary_info_response: typing.Optional[GetBinaryInfoResponse] = None
    download_chunk_response: typing.Optional[DownloadChunkResponse] = None
    binary_transfer_error: typing.Optional[BinaryTransferError] = None
    command_error: typing.Optional[SiLAError] = None
    property_error: typing.Optional[SiLAError] = None

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = cls(request_uuid="")
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                message.request_uuid = reader.read_string()
            elif field_number == 2:
                reader.expect_type(tag, WireType.LEN)
                message.unobservable_command_response = UnobservableCommandResponse.decode(reader, reader.read_uint32())
            elif field_number == 3:
                reader.expect_type(tag, WireType.LEN)
                message.observable_command_confirmation = CommandConfirmationResponse.decode(
                    reader, reader.read_uint32()
                )
            elif field_number == 4:
                reader.expect_type(tag, WireType.LEN)
                message.observable_command_execution_info = CommandExecutionResponse.decode(
                    reader, reader.read_uint32()
                )
            elif field_number == 5:
                reader.expect_type(tag, WireType.LEN)
                message.observable_command_intermediate_response = ObservableCommandResponse.decode(
                    reader, reader.read_uint32()
                )
            elif field_number == 6:
                reader.expect_type(tag, WireType.LEN)
                message.observable_command_response = ObservableCommandResponse.decode(reader, reader.read_uint32())
            elif field_number == 7:
                reader.expect_type(tag, WireType.LEN)
                message.get_fcp_affected_by_metadata_response = MetadataResponse.decode(reader, reader.read_uint32())
            elif field_number == 8:
                reader.expect_type(tag, WireType.LEN)
                message.unobservable_property_value = PropertyResponse.decode(reader, reader.read_uint32())
            elif field_number == 9:
                reader.expect_type(tag, WireType.LEN)
                message.observable_property_value = PropertyResponse.decode(reader, reader.read_uint32())
            elif field_number == 10:
                reader.expect_type(tag, WireType.LEN)
                message.create_binary_response = CreateBinaryResponse.decode(reader, reader.read_uint32())
            elif field_number == 11:
                reader.expect_type(tag, WireType.LEN)
                message.upload_chunk_response = UploadChunkResponse.decode(reader, reader.read_uint32())
            elif field_number == 12:
                reader.expect_type(tag, WireType.LEN)
                message.delete_binary_response = DeleteBinaryResponse.decode(reader, reader.read_uint32())
            elif field_number == 13:
                reader.expect_type(tag, WireType.LEN)
                message.get_binary_info_response = GetBinaryInfoResponse.decode(reader, reader.read_uint32())
            elif field_number == 14:
                reader.expect_type(tag, WireType.LEN)
                message.download_chunk_response = DownloadChunkResponse.decode(reader, reader.read_uint32())
            elif field_number == 15:
                reader.expect_type(tag, WireType.LEN)
                message.binary_transfer_error = BinaryTransferError.decode(reader, reader.read_uint32())
            elif field_number == 16:
                reader.expect_type(tag, WireType.LEN)
                message.command_error = SiLAError.decode(reader, reader.read_uint32())
            elif field_number == 17:
                reader.expect_type(tag, WireType.LEN)
                message.property_error = SiLAError.decode(reader, reader.read_uint32())
            else:
                reader.skip_type(tag & 7)

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        if self.request_uuid:
            writer.write_uint32(10).write_string(self.request_uuid)

        if self.unobservable_command_response is not None:
            self.unobservable_command_response.encode(writer, 2)

        if self.observable_command_confirmation is not None:
            self.observable_command_confirmation.encode(writer, 3)

        if self.observable_command_execution_info is not None:
            self.observable_command_execution_info.encode(writer, 4)

        if self.observable_command_intermediate_response is not None:
            self.observable_command_intermediate_response.encode(writer, 5)

        if self.observable_command_response is not None:
            self.observable_command_response.encode(writer, 6)

        if self.get_fcp_affected_by_metadata_response is not None:
            self.get_fcp_affected_by_metadata_response.encode(writer, 7)

        if self.unobservable_property_value is not None:
            self.unobservable_property_value.encode(writer, 8)

        if self.observable_property_value is not None:
            self.observable_property_value.encode(writer, 9)

        if self.create_binary_response is not None:
            self.create_binary_response.encode(writer, 10)

        if self.upload_chunk_response is not None:
            self.upload_chunk_response.encode(writer, 11)

        if self.delete_binary_response is not None:
            self.delete_binary_response.encode(writer, 12)

        if self.get_binary_info_response is not None:
            self.get_binary_info_response.encode(writer, 13)

        if self.download_chunk_response is not None:
            self.download_chunk_response.encode(writer, 14)

        if self.binary_transfer_error is not None:
            self.binary_transfer_error.encode(writer, 15)

        if self.command_error is not None:
            self.command_error.encode(writer, 16)

        if self.property_error is not None:
            self.property_error.encode(writer, 17)

        if number:
            writer.ldelim()

        return writer.finish()



import datetime

import typing_extensions as typing


def _cmp(x, y):
    return 0 if x == y else 1 if x > y else -1


def _format_offset(off):
    s = ""
    if off is not None:
        if off.days < 0:
            sign = "-"
            off = -off
        else:
            sign = "+"
        hh, mm = divmod(off, datetime.timedelta(hours=1))
        mm, ss = divmod(mm, datetime.timedelta(minutes=1))
        s += "%s%02d:%02d" % (sign, hh, mm)
        if ss or ss.microseconds:
            s += ":%02d" % ss.seconds

            if ss.microseconds:
                s += ".%06d" % ss.microseconds
    return s


class date(datetime.date):
    """Date with time zone."""

    __slots__ = "_year", "_month", "_day", "_hashcode", "_tzinfo"

    def __new__(
        cls,
        year,
        month: typing.Optional[int] = None,
        day: typing.Optional[int] = None,
        tzinfo: typing.Optional[datetime.tzinfo] = None,
    ):
        """Constructor.

        Arguments:

        year, month, day, tzinfo (required, base 1)
        """
        if isinstance(year, bytes) and len(year) == 4 and 1 <= ord(year[2:3]) <= 12:
            self = datetime.date.__new__(cls, year)  # type: ignore
            self._setstate(year, month)
            self._hashcode = -1
            return self

        self = datetime.date.__new__(cls, year=year, month=month, day=day)  # type: ignore
        self._year = year
        self._month = month
        self._day = day
        self._hashcode = -1
        _check_tzinfo_arg(tzinfo)
        self._tzinfo = tzinfo
        return self

    def replace(
        self,
        year: typing.Optional[int] = None,
        month: typing.Optional[int] = None,
        day: typing.Optional[int] = None,
        tzinfo: typing.Union[None, bool, datetime.tzinfo] = True,
    ):
        """Return a new date with new values for the specified fields."""
        if year is None:
            year = self._year
        if month is None:
            month = self._month
        if day is None:
            day = self._day

        _tzinfo: typing.Optional[datetime.tzinfo] = None
        if tzinfo is True:
            _tzinfo = self.tzinfo
        elif isinstance(tzinfo, datetime.tzinfo):
            _tzinfo = tzinfo

        return type(self)(year, month, day, _tzinfo)

    # Read-only field accessors

    @property
    def tzinfo(self) -> typing.Optional[datetime.tzinfo]:
        """timezone info object"""
        return self._tzinfo

    def utcoffset(self):
        """Return the timezone offset as timedelta positive east of UTC (negative west of
        UTC)."""
        if self._tzinfo is None:
            return None
        offset = self._tzinfo.utcoffset(datetime.datetime(self._year, self._month, self._day, tzinfo=self._tzinfo))
        return offset

    # Comparisons of date objects with other.

    def __eq__(self, other):
        if isinstance(other, date):
            return self._cmp(other) == 0
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, date):
            return self._cmp(other) != 0
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, date):
            return self._cmp(other) <= 0
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, date):
            return self._cmp(other) < 0
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, date):
            return self._cmp(other) >= 0
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, date):
            return self._cmp(other) > 0
        return NotImplemented

    def _cmp(self, other):
        assert isinstance(other, date)
        y, m, d, tz = self.year, self.month, self.day, self.tzinfo
        y2, m2, d2, tz2 = other.year, other.month, other.day, other.tzinfo

        offset1 = (tz or datetime.timezone.utc).utcoffset(None) or datetime.timedelta(0)
        offset2 = (tz2 or datetime.timezone.utc).utcoffset(None) or datetime.timedelta(0)

        return _cmp((y, m, d, -offset1), (y2, m2, d2, -offset2))

    # Pickle support.

    def _getstate(self):
        yhi, ylo = divmod(self._year, 256)
        basestate = bytes([yhi, ylo, self._month, self._day])
        if self._tzinfo is None:
            return (basestate,)
        else:
            return (basestate, self._tzinfo)

    def _setstate(self, string, tzinfo):
        if tzinfo is not None and not isinstance(tzinfo, datetime.tzinfo):
            raise TypeError("bad tzinfo state arg")

        yhi, ylo, self._month, self._day = string
        self._year = yhi * 256 + ylo
        # pylint: disable=attribute-defined-outside-init
        self._tzinfo = tzinfo

    def __reduce__(self):
        return (self.__class__, self._getstate())

    def __repr__(self):
        """Convert to formal string, for repr().

        >>> dt = datetime(2010, 1, 1)
        >>> repr(dt)
        'datetime.datetime(2010, 1, 1, 0, 0)'

        >>> dt = datetime(2010, 1, 1, tzinfo=timezone.utc)
        >>> repr(dt)
        'datetime.datetime(2010, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)'
        """
        return "%s.%s(%d, %d, %d, tzinfo=%r)" % (
            self.__class__.__module__,
            self.__class__.__qualname__,
            self._year,
            self._month,
            self._day,
            self._tzinfo,
        )

    def isoformat(self):
        s = "%04d-%02d-%02d" % (self._year, self._month, self._day)

        off = self.utcoffset()
        tz = _format_offset(off)
        if tz:
            s += tz

        return s

    __str__ = isoformat


def _check_tzinfo_arg(tz):
    if tz is not None and not isinstance(tz, datetime.tzinfo):
        raise TypeError("tzinfo argument must be None or of a tzinfo subclass")



from .command_identifier import CommandIdentifier
from .data_type_identifier import DataTypeIdentifier
from .error_identifier import ErrorIdentifier
from .feature_identifier import FeatureIdentifier
from .intermediate_response_identifier import IntermediateResponseIdentifier
from .metadata_identifier import MetadataIdentifier
from .parameter_identifier import ParameterIdentifier
from .property_identifier import PropertyIdentifier
from .response_identifier import ResponseIdentifier

__all__ = [
    "CommandIdentifier",
    "DataTypeIdentifier",
    "ErrorIdentifier",
    "FeatureIdentifier",
    "IntermediateResponseIdentifier",
    "MetadataIdentifier",
    "ParameterIdentifier",
    "PropertyIdentifier",
    "ResponseIdentifier",
]



from .deserializer import Characters, Deserializer, EndDocument, EndElement, StartElement, Token
from .parse_error import ParseError
from .serializable import Serializable
from .serializer import Serializer

__all__ = [
    "Characters",
    "Deserializer",
    "EndDocument",
    "EndElement",
    "ParseError",
    "Serializable",
    "Serializer",
    "StartElement",
    "Token",
]



import re

import typing_extensions as typing

from .command_identifier import CommandIdentifier
from .feature_identifier import identifier


class ResponseIdentifier(CommandIdentifier):
    """
    Uniquely identifies a response of a command.

    Attributes:
      response: Unique response identifier.

    Examples:
      >>> identifier = FeatureIdentifier(
      ...     "org.silastandard/core/SiLAService/v1/Command/GetFeatureDefinition/Response/FeatureDefinition"
      ... )
    """

    _pattern: typing.ClassVar[str] = f"{CommandIdentifier._pattern}/Response/(?P<response>{identifier})"
    _regex: typing.ClassVar[re.Pattern] = re.compile(_pattern, re.IGNORECASE)

    @typing.override
    @classmethod
    def create(
        cls, originator: str, category: str, feature: str, version: int, command: str, response: str
    ) -> typing.Self:
        """
        Create a new command response identifier from the given components.

        Args:
          originator: The organization who created and owns the feature.
          category: Group features into domains of application.
          feature: Unique feature identifier.
          version: The major feature version.
          command: Unique command identifier.
          response: Unique response identifier.

        Returns:
          The created command response identifier.
        """

        identifier = super().__new__(
            cls, f"{originator}/{category}/{feature}/v{version}/Command/{command}/Response/{response}"
        )
        identifier._data = {
            "originator": originator,
            "category": category,
            "feature": feature,
            "version": version,
            "command": command,
            "response": response,
        }

        return identifier

    @property
    def response(self) -> str:
        """Unique response identifier."""

        return self._data["response"]



import base64

import grpc
import grpc.aio
import typing_extensions as typing

from ..protobuf import DecodeError, EncodeError, Message, Reader, WireType, Writer


class SiLAError(Exception, Message):
    """
    Any error that can occur during communication via SiLA.

    Args:
      message: An error message providing additional context or
        details about the error and how to resolve it.
    """

    def __init__(self, message: str, *args: object) -> None:
        super().__init__(message, *args)

    @property
    def message(self) -> str:
        """An explanation of why the error occurred."""

        return self.args[0]

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> "SiLAError":
        from .defined_execution_error import DefinedExecutionError
        from .framework_error import FrameworkError
        from .undefined_execution_error import UndefinedExecutionError
        from .validation_error import ValidationError

        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = None
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.LEN)
                message = ValidationError.decode(reader, reader.read_uint32())
            elif field_number == 2:
                reader.expect_type(tag, WireType.LEN)
                message = DefinedExecutionError.decode(reader, reader.read_uint32())
            elif field_number == 3:
                reader.expect_type(tag, WireType.LEN)
                message = UndefinedExecutionError.decode(reader, reader.read_uint32())
            elif field_number == 4:
                reader.expect_type(tag, WireType.LEN)
                message = FrameworkError.decode(reader, reader.read_uint32())
            else:
                reader.skip_type(tag & 7)

        if message is None:
            msg = "Expected at least one valid error type."
            raise DecodeError(msg, reader.cur)

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        msg = "Can only encode subclasses of SiLA error."
        raise EncodeError(msg)

    async def to_rpc_error(self, context: grpc.aio.ServicerContext) -> typing.Self:
        """
        Abort the grpc handler and inform the client about the error.

        Args:
          context: The gRPC call context.

        Returns:
          The Error instance, allowing for method chaining.
        """

        message = self.encode()
        details = base64.standard_b64encode(message).decode("ascii")
        await context.abort(code=grpc.StatusCode.ABORTED, details=details)

        return self

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SiLAError):
            return NotImplemented

        return issubclass(self.__class__, other.__class__) and self.args == other.args



import asyncio
import collections.abc

import typing_extensions as typing

T = typing.TypeVar("T")


class QueueShutDown(Exception):
    """Raised when putting on to or getting from a shut-down Queue."""


class Stream(collections.abc.AsyncIterator[T]):
    """An AsyncIterable you can asynchronously add items to."""

    def __init__(self, maxsize: int = 0):
        self._queue = asyncio.Queue[typing.Optional[T]]()
        self._maxsize = maxsize
        self._closed = False

    @property
    def size(self) -> int:
        """Number of items in the queue."""

        return self._queue.qsize()

    @property
    def full(self) -> bool:
        """Whether there are `maxsize` items in the Stream."""

        if self._maxsize <= 0:
            return False

        return self.size >= self._maxsize

    def next(self, item: T) -> None:
        """
        Manually set the next item.

        Args:
          item: The item to add to the queue.

        Raises:
          QueueShutDown: If the queue has been shut down.
        """

        if self._closed:
            msg = "Cannot push to a closed queue."
            raise QueueShutDown(msg)

        self._put(item)

    def close(self) -> None:
        """Signal that no more items will be added."""

        self._closed = True
        self._put(None)

    async def get(self, timeout: typing.Optional[float] = None) -> T:
        """
        Get the next item from the queue.

        Args:
          timeout: The number of seconds to wait for. Omit to wait
            indefinitely.

        Returns:
          The next item from the queue.

        Raises:
          TimeoutError: If no new item occured within the given timeout.
        """

        try:
            item = await asyncio.wait_for(self._queue.get(), timeout)
            self._queue.task_done()

            if item is None:
                raise TimeoutError

            return item
        except (TimeoutError, asyncio.TimeoutError):
            raise TimeoutError from None

    def __aiter__(self) -> collections.abc.AsyncIterator[T]:
        return self

    async def __anext__(self) -> T:
        try:
            item = await self._queue.get()
            self._queue.task_done()

            if item is None:
                raise StopAsyncIteration

            return item
        except asyncio.CancelledError:
            raise StopAsyncIteration from None

    def _put(self, item: typing.Optional[T]) -> None:
        if item is not None and self.full:
            self._queue._queue.popleft()  # type: ignore

        self._queue.put_nowait(item)



class EncodeError(Exception):
    """
    Raised when encoding operations encounter invalid or insufficient data.

    Args:
      message: An explanation of why the encoding error occurred.
    """

    def __init__(self, message: str):
        super().__init__(message)

    @property
    def message(self) -> str:
        """An explanation of why the encoding error occurred."""

        return self.args[0]

    def __str__(self) -> str:
        return self.args[0]



import dataclasses

from ..validators import check_display_name, check_identifier
from .any import Any
from .data_type import DataType


@dataclasses.dataclass
class Element:
    """
    An element represents an entry in the SiLA structure data type.

    Attributes:
      identifier: Uniquely identifies the structure element within
        the scope of its structure data type. Uniqueness is checked
        without taking lower and upper case into account. Should be
        in pascal case.
      display_name: Human readable name of the structure element.
        Should be the identifier with spaces between separate words.
      description: Describes the use and purpose of the structure
        element.
      data_type: The SiLA data type of the structure element.
    """

    identifier: str
    display_name: str
    description: str = ""
    data_type: type[DataType] = Any

    def __post_init__(self) -> None:
        check_identifier(self.identifier)
        check_display_name(self.display_name)



import abc
import collections.abc

import typing_extensions as typing

from ..fdl import Deserializer, Serializable, Serializer, StartElement
from ..protobuf import Message
from .convertable import Convertable, T

if typing.TYPE_CHECKING:
    from ..common import Context


class DataType(Message, Convertable[T], Serializable, typing.Generic[T], metaclass=abc.ABCMeta):
    """The data type of any information exchanged between SiLA server and client."""

    @typing.override
    @classmethod
    @abc.abstractmethod
    async def from_native(cls, context: "Context", value: typing.Optional[T] = None, /) -> typing.Self: ...

    @typing.override
    @abc.abstractmethod
    async def to_native(self, context: "Context", /) -> T: ...

    @typing.override
    @abc.abstractmethod
    async def validate(self) -> typing.Self: ...

    @typing.override
    @classmethod
    @abc.abstractmethod
    def serialize(cls, serializer: Serializer) -> None: ...

    @typing.override
    @classmethod
    def deserialize(cls, deserializer: Deserializer) -> collections.abc.Generator[None, typing.Any, type["DataType"]]:
        from ..data_types import Constrained, Custom, List, Structure

        yield from deserializer.read_start_element(name="DataType")

        token = yield from deserializer.peek()

        if isinstance(token, StartElement) and token.name == "Basic":
            data_type = yield from deserializer.read(BasicType.deserialize(deserializer))
        elif isinstance(token, StartElement) and token.name == "List":
            data_type = yield from deserializer.read(List.deserialize(deserializer))
        elif isinstance(token, StartElement) and token.name == "Structure":
            data_type = yield from deserializer.read(Structure.deserialize(deserializer))
        elif isinstance(token, StartElement) and token.name == "Constrained":
            data_type = yield from deserializer.read(Constrained.deserialize(deserializer))
        elif isinstance(token, StartElement) and token.name == "DataTypeIdentifier":
            data_type = yield from deserializer.read(Custom.deserialize(deserializer))

        yield from deserializer.read_end_element(name="DataType")

        return data_type


class BasicType(typing.Generic[T], DataType[T]):
    """A predefined collection of SiLA data types without any child data type items."""

    @typing.override
    @classmethod
    def serialize(cls, serializer: Serializer) -> None:
        serializer.start_element("DataType")
        serializer.write_str("Basic", cls.__name__)
        serializer.end_element("DataType")

    @typing.override
    @classmethod
    def deserialize(cls, deserializer: Deserializer) -> collections.abc.Generator[None, typing.Any, type["BasicType"]]:
        from ..data_types import Any, Binary, Boolean, Date, Integer, Real, String, Time, Timestamp

        yield from deserializer.read_start_element("Basic")

        basic_type = yield from deserializer.read_str()

        yield from deserializer.read_end_element("Basic")

        if basic_type == "String":
            return String
        elif basic_type == "Integer":
            return Integer
        elif basic_type == "Real":
            return Real
        elif basic_type == "Boolean":
            return Boolean
        elif basic_type == "Binary":
            return Binary
        elif basic_type == "Date":
            return Date
        elif basic_type == "Time":
            return Time
        elif basic_type == "Timestamp":
            return Timestamp
        elif basic_type == "Any":
            return Any
        else:
            msg = f"Expected basic type value, received '{basic_type}'."
            raise ValueError(msg)



import dataclasses

import typing_extensions as typing

from ..protobuf import Message, Reader, WireType, Writer


@dataclasses.dataclass
class UnobservableCommandResponse(Message):
    """
    The response of the executed SiLA command.

    Attributes:
      value: The response of the command.
    """

    response: bytes = b""

    @typing.override
    @classmethod
    def decode(cls, reader: typing.Union[Reader, bytes], length: typing.Optional[int] = None) -> typing.Self:
        reader = reader if isinstance(reader, Reader) else Reader(reader)

        message = cls()
        end = reader.length if length is None else reader.cursor + length

        while reader.cursor < end:
            tag = reader.read_uint32()
            field_number = tag >> 3

            if field_number == 1:
                reader.expect_type(tag, WireType.LEN)
                message.response = reader.read_bytes()
            else:
                reader.skip_type(tag & 7)

        return message

    @typing.override
    def encode(self, writer: typing.Optional[Writer] = None, number: typing.Optional[int] = None) -> bytes:
        writer = writer or Writer()

        if number:
            writer.write_uint32((number << 3) | 2).fork()

        if self.response:
            writer.write_uint32(10).write_bytes(self.response)

        if number:
            writer.ldelim()

        return writer.finish()



import collections.abc

import typing_extensions as typing


class DecodeError(Exception):
    """
    Raised when decoding operations encounter invalid or insufficient data.

    Args:
      message: An explanation of why the decoding error occurred.
      offset: The position in the data where the error occurred.
      path: The path to the element where the error occurred.
    """

    def __init__(
        self, message: str, offset: int, path: typing.Optional[collections.abc.Sequence[typing.Union[int, str]]] = None
    ):
        super().__init__(message, offset, list(path) if path is not None else [])

    @property
    def message(self) -> str:
        """An explanation of why the decoding error occurred."""

        return self.args[0]

    @property
    def offset(self) -> int:
        """The position in the data where the error occurred."""

        return self.args[1]

    @property
    def path(self) -> collections.abc.Sequence[typing.Union[int, str]]:
        """The path to the element where the error occurred."""

        return self.args[2]

    def __str__(self) -> str:
        return self.args[0]
